@@ -5,13 +5,32 @@ from SigProfilerMatrixGenerator.scripts import SigProfilerMatrixGeneratorFunc as
 import shutil
 from numpy import median
 import pickle
+import bisect
 
+def pullVaf (project, project_path, sanger=True, TCGA=False, correction=True):
+	'''
+	Collects the VAFs from the original mutation files. Assumes that these are provided in the same 
+	format as Sanger or TCGA.
 
-def pullVaf (project, project_path, sanger=True, TCGA=False):
-	
+	Parameters:
+			 project	->	user provided project name (string)
+		project_path	->	the directory for the given project (string)
+			  sanger	->	optional parameter that informs the tool of what format the VAF scores are provided. This is required when subClassify=True (boolean; default=True)
+				TCGA	->	optional parameter that informs the tool of what format the VAF scores are provided. This is required when subClassify=True and sanger=False (boolean; default=False)
+		  correction	->	optional parameter to perform a genome-wide mutational density correction (boolean; default=False)
+
+	Returns:
+		None
+
+	Outputs:
+		New clustered mutation files that contain the VAF for each mutation.
+	'''
+	path_suffix = ''
+	if correction:
+		path_suffix += "_corrected"
 	vcf_path = project_path + "input/"
-	clusteredMutsPath = project_path + "output/vcf_files/" + project + "_clustered/"
-	clusteredMutsFile = project_path + "output/vcf_files/" + project + "_clustered" + "/SNV/" + project + "_clustered.txt"
+	clusteredMutsPath = project_path + "output/vcf_files" + path_suffix + "/" + project + "_clustered/"
+	clusteredMutsFile = project_path + "output/vcf_files" + path_suffix + "/" + project + "_clustered" + "/SNV/" + project + "_clustered.txt"
 	vcf_files = [x for x in os.listdir(vcf_path) if x != ".DS_Store"]
 
 
@@ -26,6 +45,8 @@ def pullVaf (project, project_path, sanger=True, TCGA=False):
 						continue
 					lines = lines.strip().split()
 					chrom = lines[0]
+					if len(chrom) > 3:
+						chrom = chrom[3:]
 					pos = lines[1]
 					ref = lines[3]
 					alt = lines[4]
@@ -44,9 +65,7 @@ def pullVaf (project, project_path, sanger=True, TCGA=False):
 				lines = lines.strip().split()
 				sample = lines[1]
 				newKey = ":".join([lines[5], lines[6], lines[8], lines[9]])
-
 				vaf = vafs[sample][newKey]
-
 				lines.append(str(vaf))
 				print("\t".join([x for x in lines]), file=out)
 
@@ -85,20 +104,41 @@ def pullVaf (project, project_path, sanger=True, TCGA=False):
 				print("\t".join([x for x in lines]), file=out)
 
 
+def findClustersOfClusters (project, chrom_based, project_parent_path, windowSize, chromLengths, regions, genome, imds, correction=True):
+	'''
+	Collects the VAFs from the original mutation files. Assumes that these are provided in the same 
+	format as Sanger or TCGA.
 
+	Parameters:
+				project	->	user provided project name (string)
+			chrom_based	->	option to generate IMDs per chromosome (boolean; default=False)
+	project_parent_path	->	the directory for the given project (string)
+			 windowSize	->	the window size used to calculate the mutation densities across the genome (integer; default=None)
+		   chromLengths	->	a dictionary of the cumulative chromosome lengths for a given reference genome (dictionary)
+				regions	->	a dictionary that contains all of the regions used for calculating corrected IMDs. If correction=False, then it returns an empty datastructure. (dictionary)
+				 genome	->	the reference genome used for the given analysis (string)
+				   imds	->	a dictionary of all of the corrected IMDs. If correction=False, then it returns an empty datastructure (dictionary)
+			 correction	->	optional parameter to perform a genome-wide mutational density correction (boolean; default=False)
 
+	Returns:
+		None
 
+	Outputs:
+		Subclassification files for all clustered mutations.
+	'''
+	path_suffix = ''
+	path_suffix2 = ''
+	if chrom_based:
+		path_suffix = "_chrom"
+	if correction:
+		path_suffix2 += "_corrected"
 
-
-def findClustersOfClusters (project, chrom_based, project_parent_path):
-	project_path = project_parent_path + "output/vcf_files/ESCC_clustered/"
+	project_path = project_parent_path + "output/vcf_files" + path_suffix2 + "/" + project + "_clustered/"
 	file  = project_path + project + '_clustered_vaf.txt'
 	out_file = project_path + project + '_clusters_of_clusters.txt'
 	out_file2 = project_path + project + '_clusters_of_clusters_imd.txt'
 
-	path_suffix = ''
-	if chrom_based:
-		path_suffix = "_chrom"
+
 
 	out_file3 = project_path + 'subclasses' + path_suffix + '/class1/' + project + '_clustered_class1.txt'
 	out_file4 = project_path + 'subclasses' + path_suffix + '/class2/' + project + '_clustered_class2.txt'
@@ -107,7 +147,6 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 	out_file7 = project_path + 'subclasses' + path_suffix + '/class1b/' + project + '_clustered_class1b.txt'
 	out_file8 = project_path + 'subclasses' + path_suffix + '/class1c/' + project + '_clustered_class1c.txt'
 
-
 	if os.path.exists(project_path + 'subclasses' + path_suffix + '/class1/'):
 		shutil.rmtree(project_path + 'subclasses' + path_suffix + '/class1/')
 	os.makedirs(project_path + 'subclasses' + path_suffix + '/class1/')
@@ -126,7 +165,6 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 	if os.path.exists(project_path + 'subclasses' + path_suffix + '/class1c/'):
 		shutil.rmtree(project_path + 'subclasses' + path_suffix + '/class1c/')
 	os.makedirs(project_path + 'subclasses' + path_suffix + '/class1c/')
-
 	cutoff = 10001
 	imd = 21
 	vaf_cut = 0.1
@@ -137,9 +175,11 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 	else:
 		with open(project_parent_path + "output/simulations/data/imds.pickle", "rb") as handle:
 			imdsData = pickle.load(handle) 
-
+	if correction:
+		with open(project_parent_path + "output/simulations/data/imds_corrected.pickle", "rb") as handle:
+			imds_corrected = pickle.load(handle) 	
 	first_pass = True
-	total_muts = 0
+	total_muts = {}
 	if first_pass:
 		mnv_length = 0
 		len_mnvs = {}
@@ -158,15 +198,23 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 			chrom = lines[i][5]
 			pos = int(lines[i][6])
 			samp = lines[i][1]
-
+			if samp not in total_muts:
+				total_muts[samp] = 0
 			if prev_samp == samp:
 				if prev_chrom == chrom:
-					if pos - prev_pos < cutoff:
+					# print(regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])])
+					# print(imds_corrected[samp])
+					# print(imds_corrected[samp][regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])]])
+					# if regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])] not in imds_corrected[samp]:
+					# 	print(samp, regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])])
+					# 	print(chrom, chromLengths[genome][chrom], pos)
+					# 	continue
+					if pos - prev_pos < cutoff or (correction and ((regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])] - (pos + chromLengths[genome][chrom]) < windowSize) and (pos - prev_pos) < imds_corrected[samp][regions[samp][bisect.bisect_left(regions[samp], pos + chromLengths[genome][chrom])]])):
 						distances.append(pos-prev_pos)
 						mnv_length += 1
 						lines[i-1] = [str(count)] + lines[i-1]
 						print("\t".join([x for x in lines[i-1]]), file=out)
-						total_muts += 1
+						total_muts[samp] += 1
 					else:
 						mnv_length += 1
 						if samp not in len_mnvs:
@@ -182,7 +230,7 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 						mnv_length = 0
 						lines[i-1] = [str(count)] + lines[i-1]
 						print("\t".join([x for x in lines[i-1]]), file=out)
-						total_muts += 1
+						total_muts[samp] += 1
 						count += 1
 						print("\n\n", file=out)
 				else:
@@ -201,7 +249,7 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 
 					lines[i-1] = [str(count)] + lines[i-1]
 					print("\t".join([x for x in lines[i-1]]), file=out)
-					total_muts += 1
+					total_muts[samp] += 1
 					count += 1
 					print("\n\n", file=out)
 			else:
@@ -220,7 +268,7 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 
 				lines[i-1] = [str(count)] + lines[i-1]
 				print("\t".join([x for x in lines[i-1]]), file=out)
-				total_muts += 1
+				total_muts[samp] += 1
 				count += 1
 				count = 1
 				print("\n\n################ New Sample #################", file=out)
@@ -228,8 +276,10 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 
 		lines[i] = [str(count)] + lines[i]
 		print("\t".join([x for x in lines[i]]), file=out)
-		total_muts += 1
+		total_muts[samp] += 1
 		out.close()
+		with open(project_parent_path + "output/simulations/data/originalCounts" + path_suffix2 + ".pickle", "wb") as f:
+			pickle.dump(total_muts, f)
 
 
 	if True:
@@ -239,6 +289,13 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 		lines = []
 		count = 1
 		with open(out_file) as f, open(out_file2, 'w') as out2, open(out_file3, 'w') as out3, open(out_file4, 'w') as out4, open(out_file5, 'w') as out5, open(out_file6, 'w') as out6, open(out_file7, 'w') as out7, open(out_file8, 'w') as out8:
+			print("HEADER", file=out2)
+			print("HEADER", file=out3)
+			print("HEADER", file=out4)
+			print("HEADER", file=out5)
+			print("HEADER", file=out6)
+			print("HEADER", file=out7)
+			print("HEADER", file=out8)
 			for line in f:
 				line = line.strip().split()[1:]
 				if line != []:
@@ -263,15 +320,23 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 							lines = []
 							continue
 						else:
-							# imd = imdsData[lines[0][2]]
-							distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if int(y[7])-int(x[7]) > 1])
-							vafs = len([abs(float(y[11]) - float(x[11])) for x,y in zip(lines, lines[1:]) if round(abs(float(y[11]) - float(x[11])),4) > vaf_cut])
-							unknownVafs = len([abs(float(y[11]) - float(x[11])) for x,y in zip(lines, lines[1:]) if abs(float(y[11]) - float(x[11])) < 0])
+							# print(imdsData[lines[0][1]])
+							# print(chromLengths[genome][lines[0][5]])
+							# print(regions[lines[0][1]][bisect.bisect_left(regions[lines[0][1]], (int(lines[0][7]) + chromLengths[genome][lines[0][5]]))])
+							# print((regions[lines[0][1]][bisect.bisect_left(regions[lines[0][1]], (int(lines[0][7]) + chromLengths[genome][lines[0][5]]))] - (int(lines[0][7]) + chromLengths[genome][lines[0][5]])))
+							# print(imds_corrected[lines[0][1]][regions[lines[0][1]][bisect.bisect_left(regions[lines[0][1]], int(lines[0][7]) + chromLengths[genome][lines[0][5]])]])
+							if correction:
+								distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if int(y[7])-int(x[7]) > 1 and (int(y[7])-int(x[7]) <= imdsData[y[1]] or (regions[y[1]][bisect.bisect_left(regions[y[1]], (int(y[7]) + chromLengths[genome][y[5]]))] - (int(y[7]) + chromLengths[genome][y[5]]) < windowSize and int(y[-2]) < imds_corrected[y[1]][regions[y[1]][bisect.bisect_left(regions[y[1]], int(y[7]) + chromLengths[genome][y[5]])]]))])
+								distancesFailed = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if (int(y[7])-int(x[7]) > imdsData[y[1]] and (regions[y[1]][bisect.bisect_left(regions[y[1]], (int(y[7]) + chromLengths[genome][y[5]]))] - (int(y[7]) + chromLengths[genome][y[5]]) < windowSize and int(y[-2]) < imds_corrected[y[1]][regions[y[1]][bisect.bisect_left(regions[y[1]], int(y[7]) + chromLengths[genome][y[5]])]]))])
+							else:
+								distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if int(y[7])-int(x[7]) > 1 and (int(y[7])-int(x[7]) <= imdsData[y[1]])])
+								distancesFailed = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if (int(y[7])-int(x[7]) > imdsData[y[1]])])
+							vafs = len([abs(float(y[-1]) - float(x[-1])) for x,y in zip(lines, lines[1:]) if round(abs(float(y[-1]) - float(x[-1])),4) > vaf_cut])
+							unknownVafs = len([abs(float(y[-1]) - float(x[-1])) for x,y in zip(lines, lines[1:]) if abs(float(y[-1]) - float(x[-1])) < 0])
 							distancesActual = [int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:])]
-							vafsActual = [abs(float(y[11]) - float(x[11])) for x,y in zip(lines, lines[1:])]
-							# imdsDistances = len([int(y[7])-int(x[7]) for x,y in zip(lines, lines[1:]) if int(y[7])-int(x[7]) > imd])
+							vafsActual = [abs(float(y[-1]) - float(x[-1])) for x,y in zip(lines, lines[1:])]
 
-							if unknownVafs == 0 and vafs == 0:
+							if unknownVafs == 0 and vafs == 0 and distancesFailed == 0:
 								# Class I or II
 								if distancesLine > 1:
 									writeClassII = True
@@ -379,22 +444,34 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 									for i in range(1, len(linesSubClass), 1):
 										# try:
 											if chrom_based:
-												if abs(float(linesSubClass[i][11]) - float(lineRef[11])) < vaf_cut and int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]][lineRef[5]]:
-													saveNewEvent.append(linesSubClass[i])
-													lineRef = linesSubClass[i]										
+												if correction:
+													if abs(float(linesSubClass[i][-1]) - float(lineRef[-1])) < vaf_cut and (int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]][lineRef[5]] or (regions[lineRef[1]][bisect.bisect_left(regions[lineRef[1]], (int(lineRef[7]) + chromLengths[genome][lineRef[5]]))] - (int(lineRef[7]) + chromLengths[genome][lineRef[5]]) < windowSize and int(lineRef[-2]) < imds_corrected[lineRef[1]][regions[lineRef[1]][bisect.bisect_left(regions[lineRef[1]], int(lineRef[7]) + chromLengths[genome][lineRef[5]])]])):
+														saveNewEvent.append(linesSubClass[i])
+														lineRef = linesSubClass[i]
+												else:
+													if abs(float(linesSubClass[i][-1]) - float(lineRef[-1])) < vaf_cut and int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]][lineRef[5]]:
+														saveNewEvent.append(linesSubClass[i])
+														lineRef = linesSubClass[i]																			
 											else:
-										# except:
-												if abs(float(linesSubClass[i][11]) - float(lineRef[11])) < vaf_cut and int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]]:
-													saveNewEvent.append(linesSubClass[i])
-													lineRef = linesSubClass[i]	
+												if correction:
+													if abs(float(linesSubClass[i][-1]) - float(lineRef[-1])) < vaf_cut and (int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]] or (regions[lineRef[1]][bisect.bisect_left(regions[lineRef[1]], (int(lineRef[7]) + chromLengths[genome][lineRef[5]]))] - (int(lineRef[7]) + chromLengths[genome][lineRef[5]]) < windowSize and int(lineRef[-2]) < imds_corrected[lineRef[1]][regions[lineRef[1]][bisect.bisect_left(regions[lineRef[1]], int(lineRef[7]) + chromLengths[genome][lineRef[5]])]])):
+														saveNewEvent.append(linesSubClass[i])
+														lineRef = linesSubClass[i]	
+												else:
+													if abs(float(linesSubClass[i][-1]) - float(lineRef[-1])) < vaf_cut and int(linesSubClass[i][7])-int(lineRef[7]) <= imdsData[lineRef[1]]: 
+														saveNewEvent.append(linesSubClass[i])
+														lineRef = linesSubClass[i]													
 											
 
 									if len(saveNewEvent) > 1:
-										distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if int(y[7])-int(x[7]) > 1])
-										vafs = len([abs(float(y[11]) - float(x[11])) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if round(abs(float(y[11]) - float(x[11])),4) > vaf_cut])
-										unknownVafs = len([abs(float(y[11]) - float(x[11])) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if abs(float(y[11]) - float(x[11])) < 0])
+										if correction:
+											distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if int(y[7])-int(x[7]) > 1 and (int(y[7])-int(x[7]) <= imdsData[y[1]] or (regions[y[1]][bisect.bisect_left(regions[y[1]], (int(y[7]) + chromLengths[genome][y[5]]))] - (int(y[7]) + chromLengths[genome][y[5]]) < windowSize and int(y[-2]) < imds_corrected[y[1]][regions[y[1]][bisect.bisect_left(regions[y[1]], int(y[7]) + chromLengths[genome][y[5]])]]))])
+										else:
+											distancesLine = len([int(y[7])-int(x[7]) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if int(y[7])-int(x[7]) > 1 and (int(y[7])-int(x[7]) <= imdsData[y[1]])])
+										vafs = len([abs(float(y[-1]) - float(x[-1])) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if round(abs(float(y[-1]) - float(x[-1])),4) > vaf_cut])
+										unknownVafs = len([abs(float(y[-1]) - float(x[-1])) for x,y in zip(saveNewEvent, saveNewEvent[1:]) if abs(float(y[-1]) - float(x[-1])) < 0])
 										distancesActual = [int(y[7])-int(x[7]) for x,y in zip(saveNewEvent, saveNewEvent[1:])]
-										vafsActual = [abs(float(y[11]) - float(x[11])) for x,y in zip(saveNewEvent, saveNewEvent[1:])]
+										vafsActual = [abs(float(y[-1]) - float(x[-1])) for x,y in zip(saveNewEvent, saveNewEvent[1:])]
 										# Class I or II
 										if distancesLine > 1:
 											writeClassII = True
@@ -525,37 +602,37 @@ def findClustersOfClusters (project, chrom_based, project_parent_path):
 
 	try:
 		print("Generating matrices for Class 1 mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1", "GRCh37", project_path + 'subclasses'+ path_suffix + '/class1/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1", genome, project_path + 'subclasses'+ path_suffix + '/class1/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
 	try:
 		print("Generating matrices for Class 2 mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class2", "GRCh37", project_path + 'subclasses'+ path_suffix +'/class2/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class2", genome, project_path + 'subclasses'+ path_suffix +'/class2/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
 	try:
 		print("Generating matrices for Class 3 mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class3", "GRCh37", project_path + 'subclasses'+ path_suffix +'/class3/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class3", genome, project_path + 'subclasses'+ path_suffix +'/class3/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
 	try:
 		print("Generating matrices for Class 1a mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1a", "GRCh37", project_path + 'subclasses'+ path_suffix +'/class1a/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1a", genome, project_path + 'subclasses'+ path_suffix +'/class1a/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
 	try:	
 		print("Generating matrices for Class 1b mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1b", "GRCh37", project_path + 'subclasses'+ path_suffix +'/class1b/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1b", genome, project_path + 'subclasses'+ path_suffix +'/class1b/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
 	try:
 		print("Generating matrices for Class 1c mutations:")
-		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1c", "GRCh37", project_path + 'subclasses'+ path_suffix +'/class1c/', seqInfo=True)#, plot=True)
+		matrices = matGen.SigProfilerMatrixGeneratorFunc("class1c", genome, project_path + 'subclasses'+ path_suffix +'/class1c/', seqInfo=True)#, plot=True)
 		print()
 	except:
 		pass
